using System;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// B Ʈ 
    /// </summary>
    public class BTreeNode<T>
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        ///  Ű īƮ
        /// </summary>
        public int KeyCountUsed = 0;

        /// <summary>
        /// Ű 迭
        /// </summary>
        public string[] KeyArray = new string[KEY_COUNT_PER_NODE];

        /// <summary>
        ///  迭
        /// </summary>
        public T[] ValueArray = new T[KEY_COUNT_PER_NODE];

        /// <summary>
        /// ڽ 迭
        /// </summary>
        public BTreeNode<T>[] ChildArray = new BTreeNode<T>[CHILD_COUNT_PER_NODE];

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// Ű 1/2 īƮ
        /// </summary>
        private const int KEY_HALF_COUNT = 2;

        /// <summary>
        ///  Ű īƮ
        /// </summary>
        private const int KEY_COUNT_PER_NODE = 2 * KEY_HALF_COUNT;

        /// <summary>
        ///  ڽ īƮ
        /// </summary>
        private const int CHILD_COUNT_PER_NODE = KEY_COUNT_PER_NODE + 1;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region ׸ ߰ϱ - AddItem(key, value, parentKey, parentValue, parentNode)

        /// <summary>
        /// ׸ ߰ϱ
        /// </summary>
        /// <param name="key">Ű</param>
        /// <param name="value"></param>
        /// <param name="parentKey">θ Ű</param>
        /// <param name="parentValue">θ </param>
        /// <param name="parentNode">θ </param>
        public void AddItem(string key, T value, ref string parentKey, ref T parentValue, ref BTreeNode<T> parentNode)
        {
            int spot = 0;

            while(spot < KeyCountUsed)
            {
                if(string.Compare(KeyArray[spot], key) >= 0)
                {
                    break;
                }

                spot++;
            }

            if((spot < KeyCountUsed) && (KeyArray[spot] == key))
            {
                throw new Exception($"Insert error. Cannot insert item with duplicate key value '{key}'");
            }

            if(ChildArray[0] == null)
            {
                AddItemToNode(spot, key, value, null, ref parentKey, ref parentValue, ref parentNode);
            }
            else
            {
                ChildArray[spot].AddItem(key, value, ref parentKey, ref parentValue, ref parentNode);

                if(parentNode != null)
                {
                    AddItemToNode(spot, parentKey, parentValue, parentNode, ref parentKey, ref parentValue, ref parentNode);
                }
            }
        }

        #endregion
        #region ׸ ϱ - RemoveItem(key)

        /// <summary>
        /// ׸ ϱ
        /// </summary>
        /// <param name="key">Ű</param>
        public void RemoveItem(string key)
        {
            int spot = 0;

            while(spot < KeyCountUsed)
            {
                if(string.Compare(KeyArray[spot], key) >= 0)
                {
                    break;
                }

                spot++;
            }

            if((spot < KeyCountUsed) && (KeyArray[spot] == key))
            {
                if(ChildArray[0] == null)
                {
                    RemoveItemFromNode(spot);
                }
                else
                {
                    string rightMostKey = string.Empty;

                    T rightMostValue = default(T);

                    ChildArray[spot].SwapRightMost(key, ref rightMostKey, ref rightMostValue);

                    KeyArray  [spot] = rightMostKey;
                    ValueArray[spot] = rightMostValue;

                    ChildArray[spot].RemoveItem(key);
                }
            }
            else
            {
                if(ChildArray[0] == null)
                {
                    throw new Exception($"Delete error. Cannot find item with key value '{key}'");
                }

                ChildArray[spot].RemoveItem(key);
            }

            if(ChildArray[0] != null)
            {
                if(ChildArray[spot].KeyCountUsed < KEY_HALF_COUNT)
                {
                    if((spot > 0) && (ChildArray[spot - 1].KeyCountUsed > KEY_HALF_COUNT))
                    {
                        RebalanceSiblings(ChildArray[spot - 1], ChildArray[spot], ref KeyArray[spot - 1], ref ValueArray[spot - 1]);
                    }
                    else if((spot < KEY_HALF_COUNT - 1) && (ChildArray[spot + 1].KeyCountUsed > KEY_HALF_COUNT))
                    {
                        RebalanceSiblings(ChildArray[spot], ChildArray[spot + 1], ref KeyArray[spot], ref ValueArray[spot]);
                    }
                    else
                    {
                        if(spot > 0)
                        {
                            MergeSiblings(spot - 1, spot, KeyArray[spot - 1], ValueArray[spot - 1]);
                        }
                        else
                        {
                            MergeSiblings(spot, spot + 1, KeyArray[spot], ValueArray[spot]);
                        }
                    }
                }
            }
        }

        #endregion
        #region ׸ ã - FindItem(key)

        /// <summary>
        /// ׸ ã
        /// </summary>
        /// <param name="key">Ű</param>
        /// <returns>׸</returns>
        public T FindItem(string key)
        {
            int spot = 0;

            while(spot < KeyCountUsed)
            {
                if(string.Compare(KeyArray[spot], key) >= 0)
                {
                    break;
                }

                spot++;
            }

            if((spot < KeyCountUsed) && (KeyArray[spot] == key))
            {
                return ValueArray[spot];
            }

            if(ChildArray[spot] == null)
            {
                return default(T);
            }

            return ChildArray[spot].FindItem(key);
        }

        #endregion

        #region Ʈ ڽ ߰ϱ - AddToListBox(listBox, int indent)

        /// <summary>
        /// Ʈ ڽ ߰ϱ
        /// </summary>
        /// <param name="listBox">Ʈ ڽ</param>
        /// <param name="indent"></param>
        public void AddToListBox(ListBox listBox, int indent)
        {
            for(int i = 0; i < KeyCountUsed; i++)
            {
                if(ChildArray[i] != null)
                {
                    ChildArray[i].AddToListBox(listBox, indent + 4);
                }

                listBox.Items.Add(new string(' ', indent) + ValueArray[i].ToString());
            }

            if(KeyCountUsed > 0)
            {
                if(ChildArray[KeyCountUsed] != null)
                {
                    ChildArray[KeyCountUsed].AddToListBox(listBox, indent + 4);
                }
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region   忡 ׸ ߰ϱ - AddItemInNodeWithRoom(spot, key, value, node)

        /// <summary>
        ///   忡 ׸ ߰ϱ
        /// </summary>
        /// <param name="spot"></param>
        /// <param name="key">Ű</param>
        /// <param name="value"></param>
        /// <param name="node"></param>
        private void AddItemInNodeWithRoom(int spot, string key, T value, BTreeNode<T> node)
        {
            Array.Copy(KeyArray  , spot    , KeyArray  , spot + 1, KeyCountUsed - spot);
            Array.Copy(ValueArray, spot    , ValueArray, spot + 1, KeyCountUsed - spot);
            Array.Copy(ChildArray, spot + 1, ChildArray, spot + 2, KeyCountUsed - spot);

            KeyArray[spot] = key;

            ValueArray[spot] = value;

            ChildArray[spot + 1] = node;

            KeyCountUsed++;
        }

        #endregion
        #region  иϱ - SplitNode(spot, key, value, node, praentKey, parentValue, parentNode)

        /// <summary>
        ///  иϱ
        /// </summary>
        /// <param name="spot"></param>
        /// <param name="key">Ű</param>
        /// <param name="value"></param>
        /// <param name="node"></param>
        /// <param name="praentKey">θ Ű</param>
        /// <param name="parentValue">θ </param>
        /// <param name="parentNode">θ </param>
        private void SplitNode(int spot, string key, T value, BTreeNode<T> node, ref string praentKey, ref T parentValue, ref BTreeNode<T> parentNode)
        {
            string[] keyArray = new string[KEY_COUNT_PER_NODE + 1];

            Array.Copy(KeyArray, 0, keyArray, 0, spot);

            keyArray[spot] = key;

            Array.Copy(KeyArray, spot, keyArray, spot + 1, KEY_COUNT_PER_NODE - spot);

            T[] valueArray = new T[KEY_COUNT_PER_NODE + 1];

            Array.Copy(ValueArray, 0, valueArray, 0, spot);

            valueArray[spot] = value;

            Array.Copy(ValueArray, spot, valueArray, spot + 1, KEY_COUNT_PER_NODE - spot);

            BTreeNode<T>[] nodeArray = new BTreeNode<T>[CHILD_COUNT_PER_NODE + 1];

            Array.Copy(ChildArray, 0, nodeArray, 0, spot + 1);

            nodeArray[spot + 1] = node;

            Array.Copy(ChildArray, spot + 1, nodeArray, spot + 2, KEY_COUNT_PER_NODE - spot);

            Array.Copy(keyArray  , 0, KeyArray  , 0, KEY_HALF_COUNT    );
            Array.Copy(valueArray, 0, ValueArray, 0, KEY_HALF_COUNT    );
            Array.Copy(nodeArray , 0, ChildArray, 0, KEY_HALF_COUNT + 1);

            Array.Clear(KeyArray  , KEY_HALF_COUNT    , KEY_HALF_COUNT);
            Array.Clear(ValueArray, KEY_HALF_COUNT    , KEY_HALF_COUNT);
            Array.Clear(ChildArray, KEY_HALF_COUNT + 1, KEY_HALF_COUNT);

            KeyCountUsed = KEY_HALF_COUNT;

            praentKey   = keyArray[KEY_HALF_COUNT];
            parentValue = valueArray[KEY_HALF_COUNT];
            parentNode  = new BTreeNode<T>();

            Array.Copy(keyArray  , KEY_HALF_COUNT + 1, parentNode.KeyArray  , 0, KEY_HALF_COUNT    );
            Array.Copy(valueArray, KEY_HALF_COUNT + 1, parentNode.ValueArray, 0, KEY_HALF_COUNT    );
            Array.Copy(nodeArray , KEY_HALF_COUNT + 1, parentNode.ChildArray, 0, KEY_HALF_COUNT + 1);

            parentNode.KeyCountUsed = KEY_HALF_COUNT;
        }

        #endregion
        #region 忡 ׸ ߰ϱ - AddItemToNode(spot, key, value, node, parentkey, parentValue, parentNode)

        /// <summary>
        /// 忡 ׸ ߰ϱ
        /// </summary>
        /// <param name="spot"></param>
        /// <param name="key">Ű</param>
        /// <param name="value"></param>
        /// <param name="node"></param>
        /// <param name="parentkey">θ Ű</param>
        /// <param name="parentValue">θ </param>
        /// <param name="parentNode">θ </param>
        private void AddItemToNode(int spot, string key, T value, BTreeNode<T> node, ref string parentkey, ref T parentValue, ref BTreeNode<T> parentNode)
        {
            if(KeyCountUsed < KEY_COUNT_PER_NODE)
            {
                AddItemInNodeWithRoom(spot, key, value, node);

                parentkey   = null;
                parentValue = default(T);
                parentNode  = null;
            }
            else
            {
                SplitNode(spot, key, value, node, ref parentkey, ref parentValue, ref parentNode);
            }
        }

        #endregion
        #region 忡 ׸ ϱ - RemoveItemFromNode(spot)

        /// <summary>
        /// 忡 ׸ ϱ
        /// </summary>
        /// <param name="spot"></param>
        private void RemoveItemFromNode(int spot)
        {
            Array.Copy(KeyArray  , spot + 1, KeyArray  , spot    , KeyCountUsed - spot - 1);
            Array.Copy(ValueArray, spot + 1, ValueArray, spot    , KeyCountUsed - spot - 1);
            Array.Copy(ChildArray, spot + 2, ChildArray, spot + 1, KeyCountUsed - spot - 1);

            KeyCountUsed--;

            KeyArray  [KeyCountUsed    ] = null;
            ValueArray[KeyCountUsed    ] = default(T);
            ChildArray[KeyCountUsed + 1] = null;
        }

        #endregion
        #region    ȯϱ - SwapRightMost(key, rightMostKey, rightMostValue)

        /// <summary>
        ///    ȯϱ
        /// </summary>
        /// <param name="key">Ű</param>
        /// <param name="rightMostKey">  Ű</param>
        /// <param name="rightMostValue">  </param>
        private void SwapRightMost(string key, ref string rightMostKey, ref T rightMostValue)
        {
            if(ChildArray[0] == null)
            {
                rightMostKey   = KeyArray  [KeyCountUsed - 1];
                rightMostValue = ValueArray[KeyCountUsed - 1];

                KeyArray[KeyCountUsed - 1] = key;
            }
            else
            {
                ChildArray[KeyCountUsed].SwapRightMost(key, ref rightMostKey, ref rightMostValue);
            }
        }

        #endregion
        #region ̿ ϱ - RebalanceSiblings(leftNode, rightNode, middleKey, middleValue)

        /// <summary>
        /// ̿ ϱ
        /// </summary>
        /// <param name="leftNode"> </param>
        /// <param name="rightNode"> </param>
        /// <param name="middleKey">߰ Ű</param>
        /// <param name="middleValue">߰ </param>
        private static void RebalanceSiblings(BTreeNode<T> leftNode, BTreeNode<T> rightNode, ref string middleKey, ref T middleValue)
        {
            int keyCount = leftNode.KeyCountUsed + rightNode.KeyCountUsed + 1;

            int middleIndex = leftNode.KeyCountUsed;

            string[] keyArray = new string[keyCount];

            Array.Copy(leftNode.KeyArray, 0, keyArray, 0, leftNode.KeyCountUsed);

            keyArray[middleIndex] = middleKey;

            Array.Copy(rightNode.KeyArray, 0, keyArray, middleIndex + 1, rightNode.KeyCountUsed);

            T[] valueArray = new T[keyCount];

            Array.Copy(leftNode.ValueArray, 0, valueArray, 0, leftNode.KeyCountUsed);

            valueArray[middleIndex] = middleValue;

            Array.Copy(rightNode.ValueArray, 0, valueArray, middleIndex + 1, rightNode.KeyCountUsed);

            BTreeNode<T>[] childArray = new BTreeNode<T>[keyCount + 1];

            Array.Copy(leftNode.ChildArray , 0, childArray, 0              , leftNode.KeyCountUsed  + 1);
            Array.Copy(rightNode.ChildArray, 0, childArray, middleIndex + 1, rightNode.KeyCountUsed + 1);

            int leftCount = (int)((keyCount - 1) / 2);

            Array.Copy(keyArray  , 0, leftNode.KeyArray  , 0, leftCount    );
            Array.Copy(valueArray, 0, leftNode.ValueArray, 0, leftCount    );
            Array.Copy(childArray, 0, leftNode.ChildArray, 0, leftCount + 1);

            Array.Clear(leftNode.KeyArray  , leftCount    , KEY_COUNT_PER_NODE - leftCount);
            Array.Clear(leftNode.ValueArray, leftCount    , KEY_COUNT_PER_NODE - leftCount);
            Array.Clear(leftNode.ChildArray, leftCount + 1, KEY_COUNT_PER_NODE - leftCount);

            leftNode.KeyCountUsed = leftCount;

            middleKey   = keyArray  [leftCount];
            middleValue = valueArray[leftCount];

            int rightCount = keyCount - 1 - leftCount;

            Array.Copy(keyArray  , leftCount + 1, rightNode.KeyArray  , 0, rightCount    );
            Array.Copy(valueArray, leftCount + 1, rightNode.ValueArray, 0, rightCount    );
            Array.Copy(childArray, leftCount + 1, rightNode.ChildArray, 0, rightCount + 1);

            rightNode.KeyCountUsed = rightCount;
        }

        #endregion
        #region ̿ ϱ - MergeSiblings(leftSpot, rightSpot, middleKey, middleValue)

        /// <summary>
        /// ̿ ϱ
        /// </summary>
        /// <param name="leftSpot"> </param>
        /// <param name="rightSpot"> </param>
        /// <param name="middleKey">߰ Ű</param>
        /// <param name="middleValue">߰ </param>
        private void MergeSiblings(int leftSpot, int rightSpot, string middleKey, T middleValue)
        {
            int middleIndex = ChildArray[leftSpot].KeyCountUsed;

            ChildArray[leftSpot].KeyArray  [middleIndex] = middleKey;
            ChildArray[leftSpot].ValueArray[middleIndex] = middleValue;

            Array.Copy(ChildArray[rightSpot].KeyArray  , 0, ChildArray[leftSpot].KeyArray  , middleIndex + 1, ChildArray[rightSpot].KeyCountUsed    );
            Array.Copy(ChildArray[rightSpot].ValueArray, 0, ChildArray[leftSpot].ValueArray, middleIndex + 1, ChildArray[rightSpot].KeyCountUsed    );
            Array.Copy(ChildArray[rightSpot].ChildArray, 0, ChildArray[leftSpot].ChildArray, middleIndex + 1, ChildArray[rightSpot].KeyCountUsed + 1);

            ChildArray[leftSpot].KeyCountUsed += ChildArray[rightSpot].KeyCountUsed + 1;

            Array.Copy(KeyArray  , leftSpot  + 1, KeyArray  , leftSpot , KeyCountUsed - leftSpot - 1);
            Array.Copy(ValueArray, leftSpot  + 1, ValueArray, leftSpot , KeyCountUsed - leftSpot - 1);
            Array.Copy(ChildArray, rightSpot + 1, ChildArray, rightSpot, KeyCountUsed - rightSpot   );

            KeyCountUsed--;

            KeyArray  [KeyCountUsed    ] = null;
            ValueArray[KeyCountUsed    ] = default(T);
            ChildArray[KeyCountUsed + 1] = null;
        }

        #endregion
    }
}